%%% union(in L1, in L2, out NL)
%%%		L1: a list
%%%		L2: a list
%%%		NL: the list that is obtained as the union of L1 and L2

union([],L,L).
union([E|L1],L2,NL):-
		member(E,L2),
		!,
		union(L1,L2,NL).
union([E|L1],L2,[E|NL]):-
		union(L1,L2,NL).

%%% delete_list(in L1, in L2, out NL)
%%%		L1: a list
%%%		L2: a list
%%%		NL: the list that results by removing all the elements occurring in L2 from L1

delete_list(L,[],L).
delete_list(L1,[E|L2],NL):-
		delete(L1,E,NL1),
		delete_list(NL1,L2,NL).

%%% disjoint_lists(in L1, in L2)
%%%		L1: a list
%%%		L2: a list
%%%		Action:
%%%				success		if L1 and L2 do NOT share any element
%%%				fail		if L1 and L2 DO share some elements

disjoint_lists([],_).
disjoint_lists([H|_],L):-
		member(H,L),
		!,
		fail.
disjoint_lists([_|L1],L2):-
		disjoint_lists(L1,L2).

%%% repeated_elements(in L1, out L2)
%%%		L1: a list
%%%		L2: the list containing the repeated elements in L1

repeated_elements([],[]).
repeated_elements([E|L1],[E|L2]):-
		member(E,L1),
		!,
		repeated_elements(L1,L2).
repeated_elements([_|L1],L2):-
		repeated_elements(L1,L2).

%%% union_in_ascending_order(in L1, in L2, out NL)
%%%		L1: an ordered list of numbers
%%%		L2: a list of numbers (may be not an ordered list)
%%%		NL: the ordered list that results from the union of L1 and L2

union_in_ascending_order(L,[],L).
union_in_ascending_order(L1,[E|L2],NL):-
		insert_in_ascending_order(L1,E,NL1),
		union_in_ascending_order(NL1,L2,NL).

%%% insert_in_ascending_order(in L, in E, out NL)
%%%		L: an ordered list of numbers
%%%		E: a number
%%%		NL: the ordered list that results from the inserting E into L

insert_in_ascending_order([],E,[E]).
insert_in_ascending_order([E|L],E,[E|L]):-
		!.
insert_in_ascending_order([E1|L],E2,[E2,E1|L]):-
		E1 > E2,
		!.
insert_in_ascending_order([E1|L],E2,[E1|NL]):-
		insert_in_ascending_order(L,E2,NL).

%%% getCartesianProduct(in L, in N, out NL)
%%%		L: a list of elements
%%%		N: times to apply product
%%%		NL: the result to apply Cartesian product N times to L

getCartesianProduct(_,0,[[]]):-
		!.
getCartesianProduct(L,N,NL):-
		NN is N - 1,
		getCartesianProduct(L,NN,L1),
		combineLists(L,L1,NL).

%%% combineLists(in L1, in L2, out NL)
%%%		L1, L2: lists of elements
%%%		NL: the list that result of combining each element of L1 with each element of L2

combineLists([],_,[]).
combineLists([E|L1],L2,NL):-
		combineElementList(E,L2,NL1),
		combineLists(L1,L2,NL2),
		append(NL1,NL2,NL).

%%% combineElementList(in E, in L, out NL)
%%%		E: an element
%%%		L: a list
%%%		NL: the list that is obtained from combining E with each element of L

combineElementList(_,[],[]).
combineElementList(E,[H|L],[[E|H]|NL]):-
		combineElementList(E,L,NL).

%%% min_list(in L, out N)
%%%		L: a list
%%%		N: the minimum number of the list

min_list([],0):-
		!.
min_list([N],N):-
		!.
min_list([H|L],N):-
		!,
		min_list(L,I),
		N is min(I,H).

%%% max_list(in L, out N)
%%%		L: a list
%%%		N: the maximum number of the list

max_list([],0):-
		!.
max_list([N],N):-
		!.
max_list([H|L],N):-
		!,
		max_list(L,I),
		N is max(I,H).

%%% remove_duplicates(in L, out NL)
%%%		L: a list
%%%		NL: the list that results from removing the duplicated elements from L

remove_duplicates([],[]).
remove_duplicates([E|L1],[E|L2]):-
		delete(L1,E,NL1),
		remove_duplicates(NL1,L2).

%%% no_doubles(in L)
%%%		L: a list
%%%		Action:
%%%				success		if L contains NO repeated elements
%%%				fail		if L DOES contain repeated elements

no_doubles([]).
no_doubles([H|L]):-
		member(H,L),
		!,
		fail.
no_doubles([_|L]):-
		no_doubles(L).

%%% non_member(in E, in L)
%%%		E: an element
%%%		L: a list
%%%		Action:
%%%				success		if E does NOT occur in L
%%%				fail		if E DOES occur in L

non_member(E,L):-
		member(E,L),
		!,
		fail.
non_member(_,_).

%%% sublist(in L1, in L2)
%%%		L1, L2: two lists
%%%		Action:
%%%				success		if L1 IS subset of L2
%%%				fail		if L1 is NOT subset of L2

sublist([],_).
sublist([E|L1],L2):-
		member(E,L2),
		sublist(L1,L2).

%%% remove_variables(in L, out NL)
%%%		L: a list
%%%		NL: the list that is obtained by removing all variables from L

remove_variables([],[]).
remove_variables([V|L],NL):-
		var(V),
		!,
		remove_variables(L,NL).

remove_variables([E|L],[E|NL]):-
		remove_variables(L,NL).
