% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% %																				% %
% %		Part II:		Ontology information									% %
% %																				% %
% %				All the predicates to get|set information in the ontology		% %
% %																				% %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %


% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% %																				% %
% %		Axiom Internal Representation											% %
% %																				% %
% %		Each axiom is stored using the Prolog (dynamic) predicate "axiom/6":	% %
% %																				% %
% %		axiom(Code,File,FirstLine,LastLine,AxiomKey,FOFormula)					% %
% %																				% %
% %			Code: An internal code about the axiom								% %
% %				-3 -> Documentation												% %
% %				-2 -> Empty line												% %
% %				-1 -> Comment													% %
% %				 0 -> FO axiom													% %
% %				 1 -> FO axiom with rowList										% %
% %				 2 -> FO axiom with rowOperator									% %
% %				 3 -> Domain													% %
% %				 4 -> Schema													% %
% %				 5 -> Axiom with functions or non-FO axiom						% %
% %				  -> ...														% %
% %																				% %
% %			File: name of file of the axiom										% %
% %																				% %
% %			FirstLine: First line of the axiom in the file						% %
% %																				% %
% %			LastLine: Last line of the axiom in the file						% %
% %																				% %
% %			AxiomKey: an atom starting fromt the name of the file				% %
% %							and followed by a number							% %
% %						comments and empty lines have the key "text"			% %
% %						documentations have the key "documentation"				% %
% %																				% %
% %			FO formulas are represented using the following functions:			% %
% %																				% %
% %					Exists (typed) ->		exists(LV,F)						% %
% %					For all (typed) ->		forall(LV,F)						% %
% %					Exists (non-typed) ->	'exists*'(LV,F)						% %
% %					For all (non-typed) ->	'forall*'(LV,F)						% %
% %					Conjunction ->			and(LF)								% %
% %					Disjunction ->			or(LF)								% %
% %					Row conjunction ->		@and(RV,H,S,F)						% %
% %					Row disjunction ->		@or(RV,H,S,F)						% %
% %					Implication ->			implies(F1,F2)						% %
% %					Equivalence ->			equiv(F1,F2)						% %
% %					Negation ->				not(F)								% %
% %					Equality ->				equal(T1,T2)						% %
% %																				% %
% %				where	LV is a list of Variables								% %
% %							(a Variable is a Prolog atom starting by '_'		% %
% %								or uppercase)									% %
% %						F, F1, F2 are formulas									% %
% %						LF is a list of formulas								% %
% %						T1, T2 are terms										% %
% %						RV is a Row variable									% %
% %						H is the head of a Row variable (a Variable)			% %
% %						S is the tail of a Row variable							% %
% %							(a Row variable)									% %
% %																				% %
% %			Terms are represented as Prolog terms								% %
% %																				% %
% %			Variables are represented using "'#var'/1"							% %
% %																				% %
% %			Row variables are represented using "'#rowVar'/1"					% %
% %																				% %
% %			Multi-length lists' instances are represented using "'#row'/1"		% %
% %																				% %
% %			Meta-information is represented using the following functions:		% %
% %																				% %
% %				Schema ->	$schema(C,AUXR,I,VP,F)								% %
% %				Domain ->	$domain(R,I,C)										% %
% %																				% %
% %				where	C is a class											% %
% %						AUXR,R,are relations									% %
% %						VP is a variable predicate								% %
% %						F is a formula											% %
% %						I is an index											% %
% %																				% %
% %		In addition, the Prolog (dynamic) predicate "numberOfLines/1" stores	% %
% %			the number of lines in each file and "numberOfAxioms/2" stores		% %
% %			the number of axioms in each file									% %
% %																				% %
% %		For statistical information, we use the predicate "axiomType/2":		% %
% %																				% %
% %		axiomType(AxiomKey,Type)												% %
% %																				% %
% %			AxiomKey: axiom key													% %
% %			Type:																% %
% %					'unitClause' if the formula is atomic						% %
% %					'nonAtomicAxiom' if the formula is non-atomic				% %
% %																				% %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

:- dynamic	axiom/6, numberOfLines/2, numberOfAxioms/2. 
:- dynamic	goal/5, numberOfGoals/2, goalInformation/3.
:- dynamic	numberOfTests/1, testInformation/5.
:- dynamic	axiomType/2.

%%% saveAxiom(in C, in FILE, in FL, in LL, in K, in F)
%%%		C: axiom type
%%%		FILE: axiom file
%%%		FL: first line of the axiom in FILE
%%%		LL: last line of the axiom in FILE
%%%		K: axiom key
%%%		F: axiom formula
%%%		Action: the axiom is stored

saveAxiom(C,FILE,FL,LL,K,F):-
		(
		 (
		  retract((axiom(C,FILE,FL,LL,K,F))),
		  !
		 );
		 (
		  true
		 )
		),
		assertz((axiom(C,FILE,FL,LL,K,F))),
		setAxiomType(K,F).

%%% getAxiom(in K, out C, out F)
%%%		K: axiom key
%%%		C: axiom type
%%%		F: axiom formula

getAxiom(K,C,F):-
		axiom(C,_,_,_,K,F).

%%% getCompleteAxiom(in K, out C, out FILE, out FL, out LL, out F)
%%%		K: axiom key
%%%		C: axiom type
%%%		FILE: axiom file
%%%		FL: first line of the axiom in FILE
%%%		LL: last line of the axiom in FILE
%%%		F: axiom formula

getCompleteAxiom(K,C,FILE,FL,LL,F):-
		axiom(C,FILE,FL,LL,K,F).

%%% getAxiomByFileLine(in FILE, in FL, out, C, out LL, out F)
%%%		FILE: axiom file
%%%		FL: first line of the axiom in FILE
%%%		C: axiom type
%%%		LL: last line of the axiom in FILE
%%%		K: axiom key
%%%		F: axiom formula

getAxiomByFileLine(FILE,FL,C,LL,K,F):-
		axiom(C,FILE,FL,LL,K,F).

%%% setAxiom(in K, in C, in F)
%%%		K: axiom key
%%%		C: axiom type
%%%		F: axiom formula
%%%		Action: the axiom of key K is update with C and F

setAxiom(K,C,F):-
		axiom(_,FILE,FL,LL,K,_),
		retract((axiom(_,_,_,_,K,_))),
		assertz((axiom(C,FILE,FL,LL,K,F))),
		setAxiomType(K,F).

%%% getNumberOfAxioms(in F, out N)
%%%		F: file
%%%		N: number of axioms in F

getNumberOfAxioms(F,N):-
		numberOfAxioms(F,N).

%%% setNumberOfAxioms(in F, in N)
%%%		F: file
%%%		N: number of axioms in F
%%%		Action: the number of axioms in F is saved

setNumberOfAxioms(F,N):-
		retractall((numberOfAxioms(F,_))),
		assertz((numberOfAxioms(F,N))).

%%% getNumberOfLines(in F, out N)
%%%		F: file
%%%		N: number of text lines in F

getNumberOfLines(F,N):-
		numberOfLines(F,N).

%%% setNumberOfLines(in F, in N)
%%%		F: file
%%%		N: number of text lines in F
%%%		Action: the number of text lines in F is saved

setNumberOfLines(F,N):-
		retractall((numberOfLines(F,_))),
		assertz((numberOfLines(F,N))).

%%% getNumberOfTests(out T)
%%%		T: number of tests

getNumberOfTests(T):-
		numberOfTests(T).

%%% setNumberOfTests(in T)
%%%		T: number of tests
%%%		Action: the number of tests is saved

setNumberOfTests(T):-
		retractall((numberOfTests(_))),
		assertz((numberOfTests(T))).

%%% saveGoal(in FILE, in FL, in LL, in K, in F)
%%%		FILE: goal file
%%%		FL: first line of the goal in FILE
%%%		LL: last line of the goal in FILE
%%%		K: goal key
%%%		F: goal formula
%%%		Action: the goal is stored

saveGoal(FILE,FL,LL,K,F):-
		(
		 (
		  retract((goal(FILE,FL,LL,K,F))),
		  !
		 );
		 (
		  true
		 )
		),
		assertz((goal(FILE,FL,LL,K,F))).

%%% getGoal(in K, out F)
%%%		K: goal key
%%%		F: goal formula

getGoal(K,F):-
		goal(_,_,_,K,F).

%%% getGoalFile(in K, out F, out FILE)
%%%		K: goal key
%%%		F: goal formula
%%%		FILE: goal file

getGoalFile(K,F,FILE):-
		goal(FILE,_,_,K,F).

%%% setGoal(in K,in F)
%%%		K: goal key
%%%		F: goal formula
%%%		Action: the goal of key K is update with F

setGoal(K,F):-
		goal(FILE,FL,LL,K,_),
		retract((goal(_,_,_,K,_))),
		assertz((goal(FILE,FL,LL,K,F))).

%%% getNumberOfGoals(in F, out N)
%%%		F: file
%%%		N: number of goals in F

getNumberOfGoals(F,N):-
		numberOfGoals(F,N).

%%% setNumberOfGoals(in F, out N)
%%%		F: file
%%%		N: number of goals in F
%%%		Action: the number of goals in F is saved

setNumberOfGoals(F,N):-
		retractall((numberOfGoals(F,_))),
		assertz((numberOfGoals(F,N))).

%%% setAxiomType(in K, in F)
%%%		K: axiom key
%%%		F: axiom formula
%%%		Action: the type of the axiom is stored using "axiomType/2"

setAxiomType(K,F):-
		retractall((axiomType(K,_))),
		(
		 (
		  isUnitClause(F),
		  !,
		  assertz((axiomType(K,'unitClause')))
		 );
		 (
		  assertz((axiomType(K,'nonAtomicAxiom')))
		 )
		).
		
%%% isUnitClause(in F)
%%%		F: formula
%%%		Action:
%%%				success		if the formula IS atomic
%%%				failure		if the formula is NON-atomic

isUnitClause(F):-
		F \= exists(_,_),
		F \= forall(_,_),
		F \= 'exists*'(_,_),
		F \= 'forall*'(_,_),
		F \= and(_),
		F \= or(_),
		F \= '@and'(_,_,_,_),
		F \= '@or'(_,_,_,_),
		F \= implies(_,_),
		F \= equiv(_,_),
		F \= not(_).

%%% saveAdimenSUMOAxioms

saveAdimenSUMOAxioms:-
		open('AdimenSUMOAxioms.pl',write,S),
		findall(_,(numberOfAxioms(SK,_),axiom(_,SK,_,_,AK,F),AK \= '$comment',AK \= 'text',AK \= '$documentation',axiomType(AK,T),term_to_atom(axiom(SK,AK,F,T),A),print_to_file(S,A),print_to_file(S,'.\n')),_),
		close(S).

%%% saveAdimenSUMOTests

saveAdimenSUMOTests:-
		open('AdimenSUMOTests.pl',write,S),
		findall(_,(testInformation(K,M,I,TKG1,TKG2),term_to_atom(testInformation(K,M,I,TKG1,TKG2),A),print_to_file(S,A),print_to_file(S,'.\n')),_),
		numberOfTests(T),
		term_to_atom(numberOfTests(T),AT),
		print_to_file(S,AT),
		print_to_file(S,'.\n'),
		close(S).


% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% %																				% %
% %		Index of auxiliary variables:											% %
% %																				% %
% %			We use the function "index_of_auxiliary_variables/1"				% %
% %																				% %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

:- dynamic index_of_auxiliary_variables/1.

%%% indexOfAuxiliaryVariables(out I)
%%%		I: current index of auxiliary variables

indexOfAuxiliaryVariables(I):-
		index_of_auxiliary_variables(I),
		NI is I + 1,
		retractall((index_of_auxiliary_variables(I))),
		assertz((index_of_auxiliary_variables(NI))).


% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% %																				% %
% %	Predicate and Constant information											% %
% %																				% %
% %		Relations are stored using the following (dynamic) predicate:			% %
% %																				% %
% %			inputPredicate(P,A)													% %
% %																				% %
% %			where	P is a predicate name										% %
% %						A is the arity of P										% %
% %																				% %
% %		Constants are stored using the following (dynamic) predicate:			% %
% %																				% %
% %			inputConstant(C)													% %
% %																				% %
% %			where	C is a constant name										% %
% %																				% %
% %		Further, we store the arity of row variables for each predicate			% %
% %																				% %
% %			rowArity(P,I,A)														% %
% %																				% %
% %			where	P is a predicate											% %
% %						I is the argument of the row variable					% %
% %						A is the arity of the row variable						% %
% %																				% %
% %		Finally, some arguments of predicates are related, if the same			% %
% %			row variable occurs in those arguments								% %
% %																				% %
% %			relatedArity(P1,I1,P2,I2)											% %
% %																				% %
% %			where	P1, P2 are predicates										% %
% %						I1, I2 are the arguments								% %
% %																				% %
% %		In the same way, we relate the arguments where tail row variables in	% %
% %			row operators occurs with the arguments where row variables			% %
% %																				% %
% %			relatedTailArity(P1,I1,P2,I2)										% %
% %																				% %
% %			where	P1, P2 are predicates										% %
% %						I1, I2 are the arguments								% %
% %																				% %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

:- dynamic	inputPredicate/2, inputConstant/1, rowArity/3,
		relatedArity/4, relatedTailArity/4.

%%% addInputConstant(in C)
%%%		C: a constant
%%%		Action: constant C is added

addInputConstant(C):-
		inputConstant(C),
		!.
addInputConstant(C):-
		assertz((inputConstant(C))).

%%% addInputPredicate(in P, in A)
%%%		P: predicate name
%%%		A: predicate arity
%%%		Action: predicate "P/A" is added

addInputPredicate(P,A):-
		inputPredicate(P,A),
		!.
addInputPredicate(P,A):-
		assertz((inputPredicate(P,A))).

%%% getInputConstant(in-out C)
%%%		C: a constant
%%%		Action:
%%%			success		if C IS a constant
%%%			fail		if c is NOT a constant

getInputConstant(C):-
		inputConstant(C),
		!.

%%% getPredicate(in-out P, in-out A)
%%%		P: predicate name
%%%		A: predicate arity
%%%		Action:
%%%			success		if "P/A" IS a predicate
%%%			fail		if "P/A" is NOT a predicate

getInputPredicate(P,A):-
		inputPredicate(P,A),
		!.

%%% getAllInputConstants(out LC)
%%%		LC: the list of all constants defined in the ontology

getAllInputConstants(LC):-
		findall(C,inputConstant(C),LC).

%%% getAllInputPredicates(out LP)
%%%		LP: the list of all predicates represented as pairs (S,A)
%%%			where
%%%				S:	predicate symbol
%%%				A:	predicate arity

getAllInputPredicates(LP):-
		findall((S,A),inputPredicate(S,A),LP).

% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% %																				% %
% %		List of ontologies:														% %
% %																				% %
% %			We use the (dynamic) predicate "ontologyList/1"						% %
% %																				% %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

:- dynamic	ontologyList/1, goalList/1.

:- retractall((ontologyList(_))).

:- assertz((ontologyList([predefinitions,merge,milo]))).

:- retractall((goalList(_))).

%:- assertz((goalList([goal00,goal01,goal10]))).
:- assertz((goalList([
						goal,
						multipleMapping,
						equalEvent,
						subclassEvent,
						commonSubclassEvent,
						antonymPattern1,
						antonymPattern2,
						antonymPattern3,
						agentRelation,
						instrumentRelation,
						resultRelation,
						negatedMultipleMapping,
						negatedEqualEvent,
						negatedSubclassEvent,
						negatedCommonSubclassEvent,
						negatedAntonymPattern1,
						negatedAntonymPattern2,
						negatedAntonymPattern3,
						negatedAgentRelation,
						negatedInstrumentRelation,
						negatedResultRelation
					]))).


%%% getOntologyList(out L)
%%%		L: list of ontologies

getOntologyList(L):-
		ontologyList(L).

%%% setOntologyList(in L)
%%%		L: list of ontologies
%%%		Action: the ontology list is updated to L

setOntologyList(L):-
		retractall((ontologyList(_))),
		assertz((ontologyList(L))).

%%% getGoalList(out L)
%%%		L: list of goals

getGoalList(L):-
		goalList(L).

%%% setGoalList(in L)
%%%		L: list of goals
%%%		Action: the goal list is updated to L

setGoalList(L):-
		retractall((goalList(_))),
		assertz((goalList(L))).

% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% %																				% %
% %		Delete previous information:											% %
% %																				% %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %

deletePreviousInformation:-
		retractall((axiom(_,_,_,_,_,_))),										% Delete previous information
		retractall((goal(_,_,_,_,_))),
		retractall((goalInformation(_,_,_))),
		retractall((testInformation(_,_,_,_,_))),
		retractall((inputPredicate(_,_))),
		retractall((inputConstant(_))),
		retractall((rowArity(_,_,_))),
		retractall((relatedArity(_,_,_,_))),
		retractall((relatedTailArity(_,_,_,_))),
		retractall((numberOfAxioms(_,_))),
		retractall((numberOfLines(_,_))),
		retractall((numberOfTests(_))),
		retractall((axiomType(_,_))),
		retractall((index_of_auxiliary_variables(_))),
		retractall(('&predicate'(_,_))),
		retractall(('&constant'(_))),
		retractall(('&class'(_))),
		retractall(('&object'(_))),
		retractall(('&topClass'(_))),
%		retractall(('&directProperSubclass'(_,_))),
		retractall(('&directSubclass'(_,_))),
%		retractall(('&properSubclass'(_,_))),
		retractall(('&subclass'(_,_))),
		retractall(('&directInstance'(_,_))),
		retractall(('&instance'(_,_))),
		retractall(('&disjoint'(_,_))),
		retractall(('&directSubrelation'(_,_))),
		retractall(('&subrelation'(_,_))),
		retractall(('&directSubAttribute'(_,_))),
		retractall(('&subAttribute'(_,_))).


% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% %																				% %
% %		Auxiliary information:													% %
% %																				% %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %


% % All the definitions and information that will be used by the compiler

% % Dynamic Predicates:
% %			- '&predicate'/2
% %			- '&constant'/1
% %			- '&class'/1
% %			- '&object'/1
% %			- '&topClass'/1
% %			- '&directSubclass'/2
% %			- '&subclass'/2
% %			- '&directInstance'/2
% %			- '&instance'/2
% %			- '&disjoint'/2
% %			- '&directSubrelation'/2
% %			- '&subrelation'/2
% %			- '&directSubAttribute'/2
% %			- '&subAttribute'/2

:- dynamic	'&predicate'/2, '&constant'/1, '&class'/1, '&object'/1,
%		'&topClass'/1, '&directProperSubclass'/2,'&properSubclass'/2,
		'&topClass'/1, '&directSubclass'/2,'&subclass'/2,
		'&directInstance'/2, '&instance'/2, '&disjoint'/2,
		'&directSubrelation'/2, '&subrelation'/2,
		'&directSubAttribute'/2, '&subAttribute'/2.


%%% initializeInternalInformation
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformation:-
		assertWithoutRepetition('&constant'('$class'),_),
		assertWithoutRepetition('&constant'('$object'),_),
		findall(C,axiom(_,_,_,_,_,'$topClass'(C)),LTC),
		initializeInternalInformationListOfTopClasses(LTC),
		initializeInternalInformationSubrelation,
		initializeInternalInformationSubAttribute,
%		initializeInternalInformationInheritableRelation,
		initializeInternalInformationRelation,
		initializeInternalInformationInstancesBySubAttribute,
		initializeInternalInformationPredicate,
		initializeInternalInformationDisjoint.

%%% assertWithoutRepetition(in A, out N)
%%%		A: an atom
%%%		N:
%%%			0	if A IS a new atom
%%%			1	if A is NOT a new atom
%%%		Action: internal information is stored using dynamic predicates

assertWithoutRepetition(A,1):-
		A,
		!.
assertWithoutRepetition(A,0):-
		assertz(A).

%%% initializeInternalInformationListOfTopClasses(in LTC)
%%%		LTC: a list of top classes
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationListOfTopClasses([]).
initializeInternalInformationListOfTopClasses([C|LTC]):-
		assertWithoutRepetition('&topClass'(C),_),
		initializeInternalInformationClass(C,[]),
		initializeInternalInformationListOfTopClasses(LTC).

%%% initializeInternalInformationClass(in C, in LC)
%%%		C: a class
%%%		LC: the list of super-classes of C
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationClass(C,LC):-
		assertWithoutRepetition('&class'(C),_),
		assertWithoutRepetition('&constant'(C),_),
		findall(SC,axiom(_,_,_,_,_,'$subclass'(SC,C)),LSC),
		initializeInternalInformationClassListOfSubclasses(C,LC,LSC),
		findall(I,axiom(_,_,_,_,_,'$instance'(I,C)),LI),
		initializeInternalInformationClassListOfInstances(C,LC,LI).

%%% initializeInternalInformationClassListOfSubclasses(in C, in LC, in LSC)
%%%		C: a class
%%%		LC: the list of super-classes of C
%%%		LSC: a list of subclasses of C
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationClassListOfSubclasses(_,_,[]).
initializeInternalInformationClassListOfSubclasses(C,LC,[SC|LSC]):-
		initializeInternalInformationClassSubclass(SC,C,LC),
		initializeInternalInformationClassListOfSubclasses(C,LC,LSC).

%%% initializeInternalInformationClassSubclass(in SC, in C, in LC)
%%%		SC: a subclass
%%%		C: the super-class of SC
%%%		LC: the list of super-classes of C
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationClassSubclass(SC,C,LC):-
		assertWithoutRepetition('&directSubclass'(SC,C),_),
		assertSubclasses(SC,[C|LC]),
		initializeInternalInformationClass(SC,[C|LC]).

%%% assertSubclass(in C, in LC)
%%%		C: a class
%%%		LC: a list of super-classes of C
%%%		Action: internal information is stored using dynamic predicates

assertSubclasses(_,[]).
assertSubclasses(SC,[C|LC]):-
		assertWithoutRepetition('&subclass'(SC,C),_),
		assertSubclasses(SC,LC).

%%% initializeInternalInformationClassListOfInstances(in C, in LC, in LI)
%%%		C: a class
%%%		LC: the list of super-classes of C
%%%		LI: a list of instances of C
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationClassListOfInstances(_,_,[]).
initializeInternalInformationClassListOfInstances(C,LC,[I|LI]):-
		initializeInternalInformationClassInstance(I,C,LC),
		initializeInternalInformationClassListOfInstances(C,LC,LI).

%%% initializeInternalInformationClassInstance(in I, in C, in LC)
%%%		I: an instance
%%%		C: the class of I
%%%		LC: the list of super-classes of C
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationClassInstance(I,C,LC):-
		assertWithoutRepetition('&object'(I),_),
		assertWithoutRepetition('&constant'(I),_),
		assertWithoutRepetition('&directInstance'(I,C),_),
		assertWithoutRepetition('&instance'(I,C),_),
		assertInstances(I,LC).

%%% assertInstances(in I, in LC)
%%%		I: an instance
%%%		LC: a list of classes of I
%%%		Action: internal information is stored using dynamic predicates

assertInstances(_,[]).
assertInstances(I,[C|LC]):-
		assertWithoutRepetition('&instance'(I,C),_),
		assertInstances(I,LC).

%%% initializeInternalInformationSubrelation,
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationSubrelation:-
		findall((SR,R),axiom(_,_,_,_,_,'subrelation'(SR,R)),LR),
		assertDirectSubrelation(LR),
		findall(SR,'&directSubrelation'(SR,_),LSR),
		remove_duplicates(LSR,NLSR),
		initializeInternalInformationSubrelationListOfRelations(NLSR).

%%% assertDirectSubrelation(in LR)
%%%		LR: a list of pairs (SR,R) where R is a relation and SR is a subrelation
%%%		Action: internal information is stored using dynamic predicates

assertDirectSubrelation([]).
assertDirectSubrelation([(SR,R)|LR]):-
		assertWithoutRepetition('&directSubrelation'(SR,R),_),
		assertDirectSubrelation(LR).

%%% initializeInternalInformationSubrelationListOfRelations(in LSR)
%%%		LSR: a list of (sub)-relations
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationSubrelationListOfRelations([]).
initializeInternalInformationSubrelationListOfRelations([SR|LSR]):-
		findall(R,'&directSubrelation'(SR,R),LR),
		assertSubrelationListOfRelations(SR,LR),
		initializeInternalInformationSubrelationListOfRelations(LSR).
		
%%% assertSubrelationListOfRelations(in SR, in LR)
%%%		SR: a subrelation
%%%		LR: a list of super-relations of SR
%%%		Action: internal information is stored using dynamic predicates

assertSubrelationListOfRelations(_,[]).
assertSubrelationListOfRelations(SR,[R|LR]):-
		assertWithoutRepetition('&subrelation'(SR,R),_),
		findall(NR,'&directSubrelation'(R,NR),NLR),
		assertSubrelationListOfRelations(SR,NLR),
		assertSubrelationListOfRelations(SR,LR).

%%% initializeInternalInformationSubAttribute,
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationSubAttribute:-
		findall((SA,A),axiom(_,_,_,_,_,'subAttribute'(SA,A)),LA),
		assertDirectSubAttribute(LA),
		findall(SA,'&directSubAttribute'(SA,_),LSA),
		remove_duplicates(LSA,NLSA),
		initializeInternalInformationSubAttributeListOfAttributes(NLSA).

%%% assertDirectSubAttribute(in LA)
%%%		LA: a list of pairs (SA,A) where A is an attribute and SA is a subAttribute
%%%		Action: internal information is stored using dynamic predicates

assertDirectSubAttribute([]).
assertDirectSubAttribute([(SA,A)|LA]):-
		assertWithoutRepetition('&directSubAttribute'(SA,A),_),
		assertDirectSubAttribute(LA).

%%% initializeInternalInformationSubAttributeListOfAttributes(in LSA)
%%%		LSA: a list of (sub)-attributes
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationSubAttributeListOfAttributes([]).
initializeInternalInformationSubAttributeListOfAttributes([SA|LSA]):-
		findall(A,'&directSubAttribute'(SA,A),LA),
		assertSubAttributeListOfAttributes(SA,LA),
		initializeInternalInformationSubAttributeListOfAttributes(LSA).

%%% assertSubAttributeListOfAttributes(in SA, in LA)
%%%		SA: a subAttribute
%%%		LA: a list of super-attributes of SA
%%%		Action: internal information is stored using dynamic predicates

assertSubAttributeListOfAttributes(_,[]).
assertSubAttributeListOfAttributes(SA,[A|LA]):-
		assertWithoutRepetition('&subAttribute'(SA,A),_),
		findall(NA,'&directSubAttribute'(A,NA),NLA),
		assertSubAttributeListOfAttributes(SA,NLA),
		assertSubAttributeListOfAttributes(SA,LA).

/*
%%% initializeInternalInformationInheritableRelation
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationInheritableRelation:-
		getAllSubclasses('InheritableRelation',LSC),
		initializeInternalInformationInheritableRelationListOfClasses(LSC).

%%% initializeInternalInformationInheritableRelationListOfClasses(in LSC)
%%%		LSC: a list of subclasses of InheritableRelation
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationInheritableRelationListOfClasses([]).
initializeInternalInformationInheritableRelationListOfClasses([SC|LSC]):-
		getAllInstancesOf(SC,LI),
		getAllSuperClasses(SC,LSSC),
		initializeInternalInformationInheritableRelationListOfInstances(SC,LI,LSSC),
		initializeInternalInformationInheritableRelationListOfClasses(LSC).

%%% initializeInternalInformationInheritableRelationListOfInstances(in SC, in LI, in LSC)
%%%		SC: a subclass of InheritableRelation
%%%		LI: a list of instances of SC
%%%		LSC: the list of super-classes of SC
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationInheritableRelationListOfInstances(_,[],_).
initializeInternalInformationInheritableRelationListOfInstances(C,[I|LI],LSC):-
		getAllSubrelations(I,LSR),
		getAllClassesOf(I,LC),
		initializeInternalInformationInheritableRelationListOfSubrelations(C,LSR,LSC,LC),
		initializeInternalInformationInheritableRelationListOfInstances(C,LI,LSC).

%%% initializeInternalInformationInheritableRelationListOfSubrelations(in C, in LSR, in LSC, in LC)
%%%		C: a subclass of InheritableRelation
%%%		LSR: a list of subrelations of C
%%%		LSC: the list of super-classes of SC
%%%		LC: the additional list of classes of the relations in LSR
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationInheritableRelationListOfSubrelations(_,[],_,_).
initializeInternalInformationInheritableRelationListOfSubrelations(C,[SR|LSR],LSC,LC):-
		assertWithoutRepetition('&object'(SR),_),
		assertWithoutRepetition('&constant'(SR),_),
		assertWithoutRepetition('&instance'(SR,C),_),
		assertInstances(SR,LSC),
		assertInstances(SR,LC),
		initializeInternalInformationInheritableRelationListOfSubrelations(C,LSR,LSC,LC).
*/

%%% initializeInternalInformationRelation
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationRelation:-
		getAllSubclasses('Relation',LSC),
		initializeInternalInformationRelationListOfClasses(LSC).

%%% initializeInternalInformationRelationListOfClasses(in LSC)
%%%		LSC: a list of subclasses of Relation
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationRelationListOfClasses([]).
initializeInternalInformationRelationListOfClasses([SC|LSC]):-
		getAllInstancesOf(SC,LI),
		getAllSuperClasses(SC,LSSC),
		initializeInternalInformationRelationListOfInstances(SC,LI,LSSC),
		initializeInternalInformationRelationListOfClasses(LSC).

%%% initializeInternalInformationRelationListOfInstances(in SC, in LI, in LSC)
%%%		SC: a subclass of Relation
%%%		LI: a list of instances of SC
%%%		LSC: the list of super-classes of SC
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationRelationListOfInstances(_,[],_).
initializeInternalInformationRelationListOfInstances(C,[I|LI],LSC):-
		getAllSubrelations(I,LSR),
		getAllClassesOf(I,LC),
		initializeInternalInformationRelationListOfSubrelations(C,LSR,LSC,LC),
		initializeInternalInformationRelationListOfInstances(C,LI,LSC).

%%% initializeInternalInformationRelationListOfSubrelations(in C, in LSR, in LSC, in LC)
%%%		C: a subclass of Relation
%%%		LSR: a list of subrelations of C
%%%		LSC: the list of super-classes of SC
%%%		LC: the additional list of classes of the relations in LSR
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationRelationListOfSubrelations(_,[],_,_).
initializeInternalInformationRelationListOfSubrelations(C,[SR|LSR],LSC,LC):-
		assertWithoutRepetition('&object'(SR),_),
		assertWithoutRepetition('&constant'(SR),_),
		assertWithoutRepetition('&instance'(SR,C),_),
		assertInstances(SR,LSC),
		assertInstances(SR,LC),
		initializeInternalInformationRelationListOfSubrelations(C,LSR,LSC,LC).

%%% initializeInternalInformationInstancesBySubAttribute
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationInstancesBySubAttribute:-
		findall((SA,A),'&subAttribute'(SA,A),LSA),
		initializeInternalInformationListOfSubAttributes(LSA).

%%% initializeInternalInformationListOfSubAttributes(in LSA)
%%%		LSA: a list of pairs (SA,A) where A is an attribute an SA is a subAttribute of A
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationListOfSubAttributes([]).
initializeInternalInformationListOfSubAttributes([(SA,A)|LSA]):-
		getAllClassesOf(A,LC),
		initializeInternalInformationSubAttributeListOfClasses(SA,LC),
		initializeInternalInformationListOfSubAttributes(LSA).

%%% initializeInternalInformationSubAttributeListOfClasses(in SA, in LC)
%%%		SA: a subAttribute
%%%		LC: a list of super-classes of a super-attribute of SA
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationSubAttributeListOfClasses(_,[]).
initializeInternalInformationSubAttributeListOfClasses(SA,[C|LC]):-
		assertWithoutRepetition('&object'(SA),_),
		assertWithoutRepetition('&constant'(SA),_),
		assertWithoutRepetition('&instance'(SA,C),_),
		initializeInternalInformationSubAttributeListOfClasses(SA,LC).

%%% initializeInternalInformationPredicate
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationPredicate:-
		assertPredefinedPredicates,
		assertSchemaDefinedPredicates,
		assertInstanceOfRelationPredicates,										%%% ¡¡¡¡ ONLY FOR SUMO !!!!
		assertSubrelationPredicates,											%%% ¡¡¡¡ ONLY FOR SUMO !!!!
		!.

%%% assertPredefinedPredicates
%%%		Action: internal information is stored using dynamic predicates

assertPredefinedPredicates:-
		assertz(('&predicate'('$topClass',1))),
		assertz(('&predicate'('$isObject',1))),
		assertz(('&predicate'('$isClass',1))),
		assertz(('&predicate'('$instance',2))),
		assertz(('&predicate'('$subclass',2))),
		assertz(('&predicate'('subrelation',2))),
		assertz(('&predicate'('subAttribute',2))),
		assertz(('&predicate'('$disjoint',2))),
		assertz(('&predicate'('$partition',2))),
		assertz(('&predicate'('$exhaustiveDecomposition',2))),
		assertz(('&predicate'('$disjointDecomposition',2))).

%%% assertSchemaDefinedPredicates
%%%		Action: internal information is stored using dynamic predicates

assertSchemaDefinedPredicates:-
		findall((C,S,A),axiom(_,_,_,_,_,'$schema'(C,S,A,_,_)),LS),
		assertSchemaDefinedPredicatesListOfSchemas(LS).

%%% assertSchemaDefinedPredicatesListOfSchemas(in LS)
%%%		LS: a list of tuples (C,S,A) where C is an schema class, S is an auxiliary relation and A is an arity
%%%		Action: internal information is stored using dynamic predicates

assertSchemaDefinedPredicatesListOfSchemas([]).
assertSchemaDefinedPredicatesListOfSchemas([(C,S,A)|LS]):-
		NA is A + 1,
		assertWithoutRepetition('&predicate'(S,NA),0),
		getAllInstancesOf(C,LP),
		assertListOfPredicatesArity(LP,A),
		assertSchemaDefinedPredicatesListOfSchemas(LS).
		
%%% assertInstanceOfRelationPredicates
%%%		Action: internal information is stored using dynamic predicates

assertInstanceOfRelationPredicates:-
		getAllInstancesOf('BinaryRelation',LP2),
		assertListOfPredicatesArity(LP2,2),
		getAllInstancesOf('TernaryRelation',LP3),
		assertListOfPredicatesArity(LP3,3),
		getAllInstancesOf('QuaternaryRelation',LP4),
		assertListOfPredicatesArity(LP4,4),
		getAllInstancesOf('QuintaryRelation',LP5),
		assertListOfPredicatesArity(LP5,5),
		getAllInstancesOf('VariableArityRelation',LPN),
		assertListOfPredicatesVariableArity(LPN).

%%% assertSubrelationPredicates
%%%		Action: internal information is stored using dynamic predicates

assertSubrelationPredicates:-
		getAllPredicates(LP),
		assertSubrelationPredicatesListOfPredicates(LP).

%%% assertSubrelationPredicatesListOfPredicates(in LP)
%%%		LP: a list of pairs (S,A) where S is predicate symbol and A is an arity
%%%		Action: internal information is stored using dynamic predicates

assertSubrelationPredicatesListOfPredicates([]).
assertSubrelationPredicatesListOfPredicates([(S,A)|LP]):-
		getAllSubrelations(S,LSR),
		assertListOfPredicatesArity(LSR,A),
		assertSubrelationPredicatesListOfPredicates(LP).

%%% assertListOfPredicates(in LP)
%%%		LP: a list of pairs (S,A) where S is predicate symbol and A is an arity
%%%		Action: internal information is stored using dynamic predicates

assertListOfPredicates([]).
assertListOfPredicates([(S,A)|LP]):-
		assertWithoutRepetition('&predicate'(S,A),_),
		assertListOfPredicates(LP).

%%% assertListOfPredicatesArity(in LS, in A)
%%%		LS: a list of predicate symbols
%%%		A: an arity
%%%		Action: internal information is stored using dynamic predicates

assertListOfPredicatesArity([],_).
assertListOfPredicatesArity([S|LS],A):-
		assertWithoutRepetition('&predicate'(S,A),_),
		assertListOfPredicatesArity(LS,A).

%%% assertListOfPredicatesVariableArity(in LS)
%%%		LS: a list of predicate symbols
%%%		Action: internal information is stored using dynamic predicates

assertListOfPredicatesVariableArity([]).
assertListOfPredicatesVariableArity([S|LS]):-
		inputPredicate(S,A),
		assertWithoutRepetition('&predicate'(S,A),_),
		assertListOfPredicatesVariableArity(LS).

%%% initializeInternalInformationDisjoint
%%%		Action: internal information is stored using dynamic predicates

initializeInternalInformationDisjoint:-
		findall(LC1,axiom(_,_,_,_,_,'$partition'(_,LC1)),LLC1),
		assertListOfRowDisjointClasses(LLC1),
		findall(LC2,axiom(_,_,_,_,_,'$disjointDecomposition'(_,LC2)),LLC2),
		assertListOfRowDisjointClasses(LLC2),
		findall((C1,C2),axiom(_,_,_,_,_,'$disjoint'(C1,C2)),LC3),
		assertListOfPairsOfDisjointClasses(LC3).

%%% assertListOfRowDisjointClasses(in LC)
%%%		LC: a list of disjoint classes represented using '#row'
%%%		Action: internal information is stored using dynamic predicates

assertListOfRowDisjointClasses([]).
assertListOfRowDisjointClasses(['#row'(C)|LC]):-
		assertListOfDisjointClasses(C),
		assertListOfRowDisjointClasses(LC).

%%% assertListOfDisjointClasses(in LC)
%%%		LC: a list of disjoint classes
%%%		Action: internal information is stored using dynamic predicates

assertListOfDisjointClasses([]).
assertListOfDisjointClasses([C|LC]):-
		assertListOfDisjointClassesClass(C,LC),
		assertListOfDisjointClasses(LC).

%%% assertListOfDisjointClassesClass(in C, in LC)
%%%		C: a class
%%%		LC: a list of disjoint classes
%%%		Action: internal information is stored using dynamic predicates

assertListOfDisjointClassesClass(_,[]).
assertListOfDisjointClassesClass(C1,[C2|LC]):-
		assertDisjointClasses(C1,C2),
		assertListOfDisjointClassesClass(C1,LC).

%%% assertListOfPairsOfDisjointClasses(in in LC)
%%%		LC: a list of pairs of disjoint classes
%%%		Action: internal information is stored using dynamic predicates

assertListOfPairsOfDisjointClasses([]).
assertListOfPairsOfDisjointClasses([(C1,C2)|LC]):-
		assertDisjointClasses(C1,C2),
		assertListOfPairsOfDisjointClasses(LC).

%%% assertDisjointClasses(in C1, in C2)
%%%		C1: a class
%%%		C2: a class
%%%		Action: internal information is stored using dynamic predicates

assertDisjointClasses(C1,C2):-
		(
		 (
		  '&disjoint'(C1,C2)
		 );
		 (
		  '&disjoint'(C2,C1)
		 )
		),
		!,
		print_to_screen('\n'),
		print_to_screen('Warning --> multiple definition of disjoint classes: '),
		print_to_screen(C1),
		print_to_screen(' and '),
		print_to_screen(C2),
		print_to_screen('\n').
assertDisjointClasses(C1,C2):-
		assertz(('&disjoint'(C1,C2))).

% % Query Predicates:
% %			- isPredicate/2														Nonvar
% %			- getAllPredicates/1												Var
% %			- isConstant/1														Nonvar
% %			- getAllConstants/1													Var
% %			- isClass/1															Nonvar
% %			- getAllClasess/1													Var
% %			- isObject/1														Nonvar
% %			- getAllObjects/1													Var
% %			- isTopClass/1														Nonvar
% %			- getTopClasses/1													Var
% %			- isDirectProperSubclass/2											Nonvar
% %			- isProperSubclass/2												Nonvar
% %			- getDirectProperSubclasses/2										Var
% %			- getAllProperSubclasses/2											Var
% %			- isSubclass/2														Nonvar
% %			- getAllSubclasses/2												Var
% %			- getDirectSuperClasses/2											Var
% %			- getAllSuperClasses/2												Var
% %			- isDirectInstanceOf/2												NonVar
% %			- isInstanceOf/2													NonVar
% %			- getDirectInstancesOf/2											Var
% %			- getAllInstancesOf/2												Var
% %			- getDirectClassesOf/2												Var
% %			- getAllClassesOf/2													Var
% %			- areDisjoint/2														Nonvar
% %			- getDisjointWith/2													Var
% %			- isDirectSubrelation/2												Nonvar
% %			- isSubrelation/2													Nonvar
% %			- getDirectSubrelations/2											Var
% %			- getAllSubrelations/2												Var
% %			- getAllSuperRelations/2											Var
% %			- isDirectSubAttribute/2											Nonvar
% %			- isSubAttribute/2													Nonvar
% %			- getDirectSubAttributes/2											Var
% %			- getAllSubAttributes/2												Var
% %			- getAllSuperAttributes/2											Var
% %			- domainIsA/2														Nonvar
% %			- isSchemaDefinedPredicate/2										Nonvar
% %			- getSchemaAuxiliarRelations/1										Var

%%% isPredicate(in S, in A)
%%%		S: a symbol
%%%		A: an arity
%%%		Action:
%%%			success		if "S/A" IS a predicate
%%%			fail		if "S/A" is NOT a predicate

isPredicate(S,A):-
		nonvar(S),
		nonvar(A),
		'&predicate'(S,A),
		!.
isPredicate(S,A):-
		(
		 (
		  var(S)
		 );
		 (
		  var(A)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isPredicate: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getAllPredicates(out LP)
%%%		LP: a list of predicates( pairs of the form "(S,A)")

getAllPredicates(LP):-
		findall((S,A),'&predicate'(S,A),LP).

%%% isConstant(in S)
%%%		S: a symbol
%%%		Action:
%%%			success		if "S" IS a constant
%%%			fail		if "S" is NOT a constant

isConstant(S):-
		nonvar(S),
		'&constant'(S),
		!.
isConstant(S):-
		var(S),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isConstant: variable as argument.'),
		print_to_screen('\n'),
		fail.

%%% getAllContants(out LC)
%%%		LC: a list of constants

getAllConstants(LC):-
		findall(C,'&constant'(C),LC).

%%% isClass(in C)
%%%		C: a symbol
%%%		Action:
%%%			success		if "C" IS a class
%%%			fail		if "C" is NOT a class

isClass(C):-
		nonvar(C),
		'&class'(C),
		!.
isClass(C):-
		var(C),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isClass: variable as argument.'),
		print_to_screen('\n'),
		fail.

%%% getAllClasses(out LC)
%%%		LC: a list of classes

getAllClasses(LC):-
		findall(C,'&class'(C),LC).

%%% isObject(in O)
%%%		O: a symbol
%%%		Action:
%%%			success		if "O" IS an object
%%%			fail		if "O" is NOT an object

isObject(O):-
		nonvar(O),
		'&object'(O),
		!.
isObject(O):-
		var(O),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isObject: variable as argument.'),
		print_to_screen('\n'),
		fail.

%%% getAllObjects(out LO)
%%%		LO: a list of objects

getAllObjects(LO):-
		findall(O,'&object'(O),LO).

%%% isTopClass(in C)
%%%		C: a symbol
%%%		Action:
%%%			success		if "C" IS a top class
%%%			fail		if "C" is NOT a top class

isTopClass(C):-
		nonvar(C),
		'&topClass'(C),
		!.
isTopClass(C):-
		var(C),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isTopClass: variable as argument.'),
		print_to_screen('\n'),
		fail.

%%% getAllTopClasses(out LC)
%%%		LC: a list of top classes

getAllTopClasses(LC):-
		findall(C,'&topClass'(C),LC).

%%% isSubclass(in SC, in C)
%%%		SC: a symbol
%%%		C: a symbol
%%%		Action:
%%%			success		if "SC" IS a subclass of "C"
%%%			fail		if "SC" is NOT a subclass of "C"

isSubclass(SC,C):-
		nonvar(SC),
		nonvar(C),
		(
		 (
		  SC = C,
		  isClass(C)
		 );
		 (
		  '&subclass'(SC,C)
		 )
		),
		!.
isSubclass(SC,C):-
		(
		 (
		  var(SC)
		 );
		 (
		  var(C)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isSubclass: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getAllSubclasses(in C, out LC)
%%%		C: a class
%%%		LC: a list of subclasses

getAllSubclasses(C,LC):-
		(
		 (
		  isClass(C),
		  !,
		  findall(SC,'&subclass'(SC,C),LPC),
		  LC = [C|LPC]
		 );
		 (
		  LC = []
		 )
		).

%%% getDirectSuperClasses(in C, out LC)
%%%		C: a class
%%%		LC: a list of direct super classes

getDirectSuperClasses(C,LC):-
		findall(SC,'&directSubclass'(C,SC),LC).

%%% getAllSuperClasses(in C, out LC)
%%%		C: a class
%%%		LC: a list of super classes

getAllSuperClasses(C,LC):-
		findall(SC,'&subclass'(C,SC),LC).

%%% isDirectInstanceOf(in I, in C)
%%%		I: a symbol
%%%		C: a symbol
%%%		Action:
%%%			success		if "I" IS a direct instance of "C"
%%%			fail		if "I" is NOT a direct instance of "C"

isDirectInstanceOf(I,C):-
		nonvar(I),
		nonvar(C),
		'&directInstance'(I,C),
		!.
isDirectInstanceOf(I,C):-
		(
		 (
		  var(I)
		 );
		 (
		  var(C)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isDirectInstanceOf: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getDirectInstancesOf(in C, out LI)
%%%		C: a class
%%%		LC: a list of direct instances

getDirectInstancesOf(C,LI):-
		findall(I,'&directInstance'(I,C),LI).

%%% isInstanceOf(in I, in C)
%%%		I: a symbol
%%%		C: a symbol
%%%		Action:
%%%			success		if "I" IS a instance of "C"
%%%			fail		if "I" is NOT a instance of "C"

isInstanceOf(I,C):-
		nonvar(I),
		nonvar(C),
		'&instance'(I,C),
		!.
isInstanceOf(I,C):-
		(
		 (
		  var(I)
		 );
		 (
		  var(C)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isInstanceOf: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getAllInstancesOf(in C, out LI)
%%%		C: a class
%%%		LC: a list of instances

getAllInstancesOf(C,LI):-
		findall(I,'&instance'(I,C),LI).

%%% getDirectClassesOf(in I, out LC)
%%%		I: an instance
%%%		LC: a list of direct classes

getDirectClassesOf(I,LC):-
		findall(C,'&directInstance'(I,C),LC).

%%% getAllClassesOf(in I, out LC)
%%%		I: an instance
%%%		LC: a list of classes

getAllClassesOf(I,LC):-
		findall(C,'&instance'(I,C),LC).

%%% areDisjoint(in C1, in C2)
%%%		C1: a symbol
%%%		C2: a symbol
%%%		Action:
%%%			success		if "C1" and "C2" ARE disjoint
%%%			fail		if "C1" and "C2" are NOT disjoint

areDisjoint(C1,C2):-
		nonvar(C1),
		nonvar(C2),
		(
		 (
		  '&disjoint'(C1,C2)
		 );
		 (
		  '&disjoint'(C2,C1)
		 )
		),
		!.
areDisjoint(C1,C2):-
		(
		 (
		  var(C1)
		 );
		 (
		  var(C2)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of areDisjoint: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getDisjointWith(in C, out LC)
%%%		C: a class
%%%		LC: a list of classes

getDisjointWith(C,LC):-
		findall(NC,('&disjoint'(C,NC) ; '&disjoint'(NC,C)),LC).

%%% isDirectSubrelation(in SR, in R)
%%%		SR: a symbol
%%%		R: a symbol
%%%		Action:
%%%			success		if "SR" IS a direct subrelation of "R"
%%%			fail		if "SR" is NOT a direct subrelation of "R"

isDirectSubrelation(SR,R):-
		nonvar(SR),
		nonvar(R),
		'&directSubrelation'(SR,R),
		!.
isDirectSubrelation(SR,R):-
		(
		 (
		  var(SR)
		 );
		 (
		  var(R)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isDirectSubrelation: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getDirectSubrelations(in R, out LR)
%%%		R: a relation
%%%		LR: a list of direct subrelations

getDirectSubrelations(R,LR):-
		findall(SR,'&directSubrelation'(SR,R),LR).

%%% isSubrelation(in SR, in R)
%%%		SR: a symbol
%%%		R: a symbol
%%%		Action:
%%%			success		if "SR" IS a subrelation of "R"
%%%			fail		if "I" is NOT a subrelation of "R"

isSubrelation(SR,R):-
		nonvar(SR),
		nonvar(R),
		'&subrelation'(SR,R),
		!.
isSubrelation(SR,R):-
		(
		 (
		  var(SR)
		 );
		 (
		  var(R)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isSubrelation: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getAllSubrelations(in R, out LR)
%%%		R: a relation
%%%		LR: a list of subrelations

getAllSubrelations(R,LR):-
		findall(SR,'&subrelation'(SR,R),LR).

%%% getAllSuperRelations(in SR, out LR)
%%%		SR: a relation
%%%		LR: a list of super-relations

getAllSuperRelations(SR,LR):-
		findall(R,'&subrelation'(SR,R),LR).

%%% isDirectSubAttribute(in SA, in A)
%%%		SA: a symbol
%%%		A: a symbol
%%%		Action:
%%%			success		if "SA" IS a direct subrAttribute of "A"
%%%			fail		if "SA" is NOT a direct subAttribute of "A"

isDirectSubAttribute(SA,A):-
		nonvar(SA),
		nonvar(A),
		'&directSubAttribute'(SA,A),
		!.
isDirectSubAttribute(SA,A):-
		(
		 (
		  var(SA)
		 );
		 (
		  var(A)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isDirectSubAttribute: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getDirectSubAttributes(in A, out LA)
%%%		A: an attribute
%%%		LA: a list of direct subAttributes

getDirectSubAttributes(A,LA):-
		findall(SA,'&directSubAttribute'(SA,A),LA).

%%% isSubAttribute(in SA, in A)
%%%		SA: a symbol
%%%		A: a symbol
%%%		Action:
%%%			success		if "SR" IS a subAttribute of "A"
%%%			fail		if "I" is NOT a subAttribute of "A"

isSubAttribute(SA,A):-
		nonvar(SA),
		nonvar(A),
		'&subAttribute'(SA,A),
		!.
isSubAttribute(SA,A):-
		(
		 (
		  var(SA)
		 );
		 (
		  var(A)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isSubAttribute: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getAllSubAttributes(in A, out LA)
%%%		A: an attribute
%%%		LA: a list of subAttributes

getAllSubAttributes(A,LA):-
		findall(SA,'&subAttribute'(SA,A),LA).

%%% getAllSuperAttributes(in SA, out LA)
%%%		SA: an attribute
%%%		LA: a list of super-Attributes

getAllSuperAttributes(SA,LA):-
		findall(A,'&subAttribute'(SA,A),LA).

%%% domainIsA(in O, in C)
%%%		O: an object
%%%		C: a class
%%%		Action:
%%%			success		if O IS an (meta-)instance of C
%%%			fail		if O is NOT an (meta-)instance of C

domainIsA(O,'$class'):-
		nonvar(O),
		!,
		isClass(O).
domainIsA(O,'$object'):-
		nonvar(O),
		!,
		isObject(O).
domainIsA(O,C):-
		nonvar(O),
		nonvar(C),
		!,
		isInstanceOf(O,C).
domainIsA(O,C):-
		(
		 (
		  var(O)
		 );
		 (
		  var(C)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of domainIsA: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% domainClass(in C1, in C2)
%%%		C1: a class
%%%		C2: a class
%%%		Action:
%%%			success		if C1 IS a subclass of C2
%%%			fail		if C1 is NOT a subclass of C2

domainClass(C1,C2):-
		nonvar(C1),
		nonvar(C2),
		!,
		isSubclass(C1,C2).

%%% isSchemaDefinedPredicate(in S, in A)
%%%		S: a symbol
%%%		A: an arity
%%%		Action:
%%%			success		if "S/A" IS an schema defined predicate
%%%			fail		if "S/A" is NOT an schema defined predicate

isSchemaDefinedPredicate(S,A):-
		nonvar(S),
		nonvar(A),
		axiom(_,_,_,_,_,'$schema'(C,_,A,_,_)),
		isInstanceOf(S,C),
		!.
isSchemaDefinedPredicate(S,A):-
		(
		 (
		  var(S)
		 );
		 (
		  var(A)
		 )
		),
		print_to_screen('\n'),
		print_to_screen('Incorrect use of isSchemaDefinedPredicate: variable(s) as argument(s).'),
		print_to_screen('\n'),
		fail.

%%% getSchemaAuxiliarRelations(out LSA)
%%%		LSA: list of auxiliar relations used by "schema/5"

getSchemaAuxiliarRelations(LSA):-
		findall((S,A),(axiom(_,_,_,_,_,'$schema'(_,S,I,_,_)),A is I + 1),LSA).

%%% getDomain(in S, in I, out T, out R)
%%%		S: a symbol
%%%		I: an index
%%%		T: the type of the i-th argument of S
%%%		R: a flag that states if the argument is a multi-length list (or row) or not
%%%			if	R = 0	then the argument is NOT a multi-length list (neither row)
%%%				R = 1	then the argument IS a multi-length list (or row)

getDomain(S,I,T,R):-
		axiom(_,_,_,_,_,'$domain'(S,I,C)),
		(
		 (
		  C = '#row'(T),
		  R = 1
		 );
		 (
		  C \= '#row'(_),
		  T = C,
		  R = 0
		 )
		),
		!.
getDomain(S,I,T,R):-															%%% ONLY FOR SUMO: domain of subrelations
		getAllSuperRelations(S,LSR),
		getDomainSubrelation(LSR,I,T,R).

%%% getDomainSubrelation(in LS, in I, out T, out R)
%%%		LS: a list of symbols
%%%		I: an index
%%%		T: the type of the i-th argument of a symbol in LS
%%%		R: a flag that states if the argument is a multi-length list (or row) or not
%%%			if	R = 0	then the argument is NOT a multi-length list (neither row)
%%%				R = 1	then the argument IS a multi-length list (or row)

getDomainSubrelation([S|_],I,T,R):-
		getDomain(S,I,T,R),
		!.
getDomainSubrelation([_|LS],I,T,R):-
		getDomainSubrelation(LS,I,T,R).

%%% getDomainSubclass(in S, in I, out T, out R)									¡¡¡¡ ONLY FOR SUMO !!!!
%%%		S: a symbol
%%%		I: an index
%%%		T: the type of the i-th argument of S
%%%		R: a flag that states if the argument is a multi-length list (or row) or not
%%%			if	R = 0	then the argument is NOT a multi-length list (neither row)
%%%				R = 1	then the argument IS a multi-length list (or row)

getDomainSubclass(S,I,T,R):-
		axiom(_,_,_,_,_,'$domainSubclass'(S,I,C)),
		(
		 (
		  C = '#row'(T),
		  R = 1
		 );
		 (
		  C \= '#row'(_),
		  T = C,
		  R = 0
		 )
		),
		!.
getDomainSubclass(S,I,T,R):-													%%% ONLY FOR SUMO: domain of subrelations
		getAllSuperRelations(S,LSR),
		getDomainSubclassSubrelation(LSR,I,T,R).

%%% getDomainSubclassSubrelation(in LS, in I, out T, out R)
%%%		LS: a list of symbols
%%%		I: an index
%%%		T: the type of the i-th argument of a symbol in LS
%%%		R: a flag that states if the argument is a multi-length list (or row) or not
%%%			if	R = 0	then the argument is NOT a multi-length list (neither row)
%%%				R = 1	then the argument IS a multi-length list (or row)

getDomainSubclassSubrelation([S|_],I,T,R):-
		getDomainSubclass(S,I,T,R),
		!.
getDomainSubclassSubrelation([_|LS],I,T,R):-
		getDomainSubclassSubrelation(LS,I,T,R).

%%% getRange(in S, out T, out R)
%%%		S: a symbol
%%%		T: the type of function S
%%%		R: a flag that states if the argument is a multi-length list (or row) or not
%%%			if	R = 0	then the argument is NOT a multi-length list (neither row)
%%%				R = 1	then the argument IS a multi-length list (or row)

getRange(S,T,R):-
		axiom(_,_,_,_,_,'$range'(S,C)),
		(
		 (
		  C = '#row'(T),
		  R = 1
		 );
		 (
		  C \= '#row'(_),
		  T = C,
		  R = 0
		 )
		),
		!.
getRange(S,T,R):-																%%% ONLY FOR SUMO: range of subrelations
		getAllSuperRelations(S,LSR),
		getRangeSubrelation(LSR,T,R).

%%% getRangeSubrelation(in LS, out T, out R)
%%%		LS: a list of symbols
%%%		T: the type of a function in LS
%%%		R: a flag that states if the argument is a multi-length list (or row) or not
%%%			if	R = 0	then the argument is NOT a multi-length list (neither row)
%%%				R = 1	then the argument IS a multi-length list (or row)

getRangeSubrelation([S|_],T,R):-
		getRange(S,T,R),
		!.
getRangeSubrelation([_|LS],T,R):-
		getRangeSubrelation(LS,T,R).

%%% getRangeSubclass(in S, out T, out R)										¡¡¡¡ ONLY FOR SUMO !!!!
%%%		S: a symbol
%%%		T: the type of function S
%%%		R: a flag that states if the argument is a multi-length list (or row) or not
%%%			if	R = 0	then the argument is NOT a multi-length list (neither row)
%%%				R = 1	then the argument IS a multi-length list (or row)

getRangeSubclass(S,T,R):-
		axiom(_,_,_,_,_,'$rangeSubclass'(S,C)),
		(
		 (
		  C = '#row'(T),
		  R = 1
		 );
		 (
		  C \= '#row'(_),
		  T = C,
		  R = 0
		 )
		),
		!.
getRangeSubclass(S,T,R):-														%%% ONLY FOR SUMO: domain of subrelations
		getAllSuperRelations(S,LSR),
		getRangeSubclassSubrelation(LSR,T,R).

%%% getRangeSubclassSubrelation(in LS, out T, out R)
%%%		LS: a list of symbols
%%%		T: the type of a function in LS
%%%		R: a flag that states if the argument is a multi-length list (or row) or not
%%%			if	R = 0	then the argument is NOT a multi-length list (neither row)
%%%				R = 1	then the argument IS a multi-length list (or row)

getRangeSubclassSubrelation([S|_],T,R):-
		getRangeSubclass(S,T,R),
		!.
getRangeSubclassSubrelation([_|LS],T,R):-
		getRangeSubclassSubrelation(LS,T,R).

%%% isSingleVariable(in T)
%%%		T: a term
%%%		Action:
%%%			success		if T IS a single variable
%%%			fail		if T is NOT a single variale

isSingleVariable('#var'(_)).

%%% isRowVariable(in T)
%%%		T: a term
%%%		Action:
%%%			success		if T IS a row variable
%%%			fail		if T is NOT a row variale

isRowVariable('#rowVar'(_)).

%%% isAnonymousRowVariable(in T)
%%%		T: a term
%%%		Action:
%%%			success		if T IS an anonymous row variable
%%%			fail		if T is NOT an anonymous row variale

isAnonymousRowVariable('#rowVar'('_')).

%%% isMultiLengthList(in T, out LT)
%%%		T: a term
%%%		LT: the elements in the list
%%%		Action:
%%%			success		if T IS a multi-length list
%%%			fail		if T is NOT a multi-length list

isMultiLengthList('#row'(L),L).

%%% toMultiLengthList(in T, out LT)
%%%		LT: a list of elements
%%%		L: the multi-length list that is obtained from LT

toMultiLengthList(L,'#row'(L)).

%%% isVarPredicate(in T)
%%%		T: a term
%%%		Action:
%%%			success		if T IS a predicate variable
%%%			fail		if T is NOT a predicate variable

isVarPredicate(VP):-
		atom_codes(VP,[C|_]),
		char_code('!',C).

%%% assertRowArity(in P, in I, in A)
%%%		P: a predicate
%%%		I: index of an argument of P
%%%		A: arity of a row list that occurs in the I-th argument of P
%%%		Action: we assert the information about the row variable if it is not available yet

assertRowArity(P,I,A):-
		rowArity(P,I,A),
		!.
assertRowArity(P,I,A):-
		assertz((rowArity(P,I,A))).

%%% assertRelatedArity(in P1, in I1, in P2, in I2)
%%%		P1,P2: predicates
%%%		I1,I2: index of arguments of P1 and P2 respectively
%%%		Action: we assert the information about the related arity of P1 and P2 if it is not available yet

assertRelatedArity(P1,I1,P2,I2):-
		relatedArity(P1,I1,P2,I2),
		!.
assertRelatedArity(P1,I1,P2,I2):-
		assertz((relatedArity(P1,I1,P2,I2))).

%%% assertRelatedTailArity(in P1, in I1, in P2, in I2)
%%%		P1,P2: predicates
%%%		I1,I2: index of arguments of P1 and P2 respectively
%%%		Action: we assert the information about the related tail arity of P1 and P2 if it is not available yet

assertRelatedTailArity(P1,I1,P2,I2):-
		relatedTailArity(P1,I1,P2,I2),
		!.
assertRelatedTailArity(P1,I1,P2,I2):-
		assertz((relatedTailArity(P1,I1,P2,I2))).

%%% addDomainAxioms(in P1, in P2, in N)
%%%		P1: a predicate that may contain row variables or mult-length lists in its arguments
%%%		P2: the corresponding predicate without row variables and multi-length lists
%%%		N: arity of P2
%%%		Action: the domain axioms of P2 are added (if not already defined)

addDomainAxioms(_,P,N):-
		findall('$domain'(P,_,_),axiom(_,_,_,_,_,'$domain'(P,_,_)),L),
		length(L,N),
		!.
addDomainAxioms(P,NP,N):-
		findall('$domain'(P,A,C),axiom(_,_,_,_,_,'$domain'(P,A,C)),L),
		sortDomainAxioms(L,NL),
		addDomainAxiomsFromList(NL,NP,N).

%%% sortDomainAxioms(in L, out NL)
%%%		L: a list of domain axioms
%%%		NL: the list that is obtained from L by sorting domain axioms in ascending order

sortDomainAxioms([],[]).
sortDomainAxioms(['$domain'(P,A,C)|L],['$domain'(NP,NA,NC)|NL]):-
		firstDomainAxiom('$domain'(P,A,C),L,'$domain'(NP,NA,NC),RL),
		sortDomainAxioms(RL,NL).

%%% firstDomainAxiom(in D, in L, out ND, out NL)
%%%		D: the current first domain axiom
%%%		L: a list of domain axioms
%%%		ND: the first domain axiom in [D|L]
%%%		NL: the remaining domain axioms in [D|L]

firstDomainAxiom(D,[],D,[]).
firstDomainAxiom('$domain'(P1,A1,C1),['$domain'(P2,A2,C2)|L],ND,['$domain'(P1,A1,C1)|NL]):-
		A2 < A1,
		!,
		firstDomainAxiom('$domain'(P2,A2,C2),L,ND,NL).
firstDomainAxiom(D1,[D2|L],ND,[D2|NL]):-
		firstDomainAxiom(D1,L,ND,NL).

%%% addDomainAxiomsFromList(in L, in P, in A)
%%%		L: a list of domain axioms
%%%		P: a predicate
%%%		A: arity of P

addDomainAxiomsFromList([],_,_):-
		print_to_screen('Error: addDomainAxioms --> End of list without finding row type.'),
		fail.
addDomainAxiomsFromList(['$domain'(_,I,'#row'(T))|L],P,A):-
		!,
		length(L,N),
		NI is A - N,
		addDomainAxiomsRowType(P,I,NI,T),
		NNI is NI + 1,
		addDomainAxiomsFromRestOfList(L,P,NNI,A).
addDomainAxiomsFromList(['$domain'(_,I,T)|L],P,A):-
		atom_concat(domain,P,K1),
		number_codes(I,LC),
		atom_codes(K2,LC),
		atom_concat(K1,K2,NK),
		assertz((axiom(3,domain,0,0,NK,'$domain'(P,I,T)))),
		addDomainAxiomsFromList(L,P,A).

%%% addDomainAxiomsRowType(in P, in I, in N, in T)
%%%		P: a predicate
%%%		I: index of the next argument to be treated
%%%		N: index of the last argument to be treated
%%%		T: types of arguments

addDomainAxiomsRowType(_,I,N,_):-
		I > N,
		!.
addDomainAxiomsRowType(P,I,N,T):-
		atom_concat(domain,P,K1),
		number_codes(I,LC),
		atom_codes(K2,LC),
		atom_concat(K1,K2,NK),
		assertz((axiom(3,domain,0,0,NK,'$domain'(P,I,T)))),
		NI is I + 1,
		addDomainAxiomsRowType(P,NI,N,T).

%%% addDomainAxiomsFromRestOfList(in L, in P, in I, in A)
%%%		L: a list of domain axioms
%%%		P: a predicate
%%%		I: next index of argument
%%%		A: arity of P

addDomainAxiomsFromRestOfList([],_,I,A):-
		I > A,
		!.
addDomainAxiomsFromRestOfList(['$domain'(_,_,T)|L],P,I,A):-
		atom_concat(domain,P,K1),
		number_codes(I,LC),
		atom_codes(K2,LC),
		atom_concat(K1,K2,NK),
		assertz((axiom(3,domain,0,0,NK,'$domain'(P,I,T)))),
		NI is I + 1,
		addDomainAxiomsFromRestOfList(L,P,NI,A).
